from typing import Any, Dict, List, cast

import yaml


class CustomDumper(yaml.SafeDumper):
    """Custom dumper that represents multi-line strings using '|'."""


def is_simple_string_list(lst: List[Any]) -> bool:
    return all(isinstance(item, str) and "\n" not in item for item in lst)


def represent_list(dumper, data) -> Any:  # type: ignore
    """Represents the list of strings as a single line if possible."""
    if is_simple_string_list(data):
        return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)
    else:
        return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=False)


def represent_ordered_dict(dumper, data) -> Any:  # type: ignore
    """Custom representer for dictionaries to maintain the insertion order."""
    return dumper.represent_mapping("tag:yaml.org,2002:map", data.items())


def multi_line_string_representer(dumper, data) -> Any:  # type: ignore
    """Represents multi-line strings with the '|' character."""
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


# Add the custom representer to only the CustomSafeDumper
CustomDumper.add_representer(list, represent_list)
CustomDumper.add_representer(dict, represent_ordered_dict)
CustomDumper.add_representer(str, multi_line_string_representer)


def yaml_dump(data: Any, **kwargs) -> str:  # type: ignore
    """YAML dump function that uses the multiline friendly SafeDumper."""
    return yaml.dump(data, Dumper=CustomDumper, **kwargs)  # type: ignore


def yaml_load(data: str) -> Dict[str, Any]:
    """YAML safe load from the file content."""
    return cast(Dict[str, Any], yaml.safe_load(data))



from __future__ import annotations

import abc
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from ._filter import Filter
from ._models import Distance, Document, Embedding, Node, Prompt, PromptStream, SerializedInput, SupportsEmbedding
from ._plugin import PluginManager, Service


# region prompt template
#######################################################################################################################
class PromptTemplate(abc.ABC):
    """Abstract base class for prompt templates.

    A prompt template provides a template for generating prompts.
    """

    @property
    @abc.abstractmethod
    def template(self) -> str:
        """The raw template represented as string."""

    @property
    @abc.abstractmethod
    def metadata(self) -> Dict[str, Any]:
        """All the metadata associated with the PromptTemplate."""

    @property
    @abc.abstractmethod
    def vars(self) -> Dict[str, Type]:
        """All the variables required by PromptTemplate."""

    @property
    @abc.abstractmethod
    def format(self) -> str:
        """The type of PromptTemplate."""

    @abc.abstractmethod
    def to_prompts(self, **kwargs: Dict[str, Any]) -> List[Prompt]:
        """Generate prompts from the template passing the variables as dict.

        Args:
            **kwargs (Dict[str, Any]): variable names as key, and replacement as value for rendering the template

        Returns:
            List[Prompt]: list of prompts generated from the template
        """


# endregion
# region prompt source
#######################################################################################################################
class PromptSource(abc.ABC):
    """Abstract base class for prompt sources.

    A prompt source provides a browsable/searchable interface for prompt templates.
    """

    @abc.abstractmethod
    def find(self, filter: Filter, stream: bool = False) -> Union[List[PromptTemplate], Iterator[PromptTemplate]]:
        """Find a prompt template for given tags.

        Args:
            filter (:class:`~bodhilib.Filter`): filter to apply on the prompt templates
            stream (Optional[bool]): option to stream the prompt templates as they are ready.
                If True, finds the prompt templates lazily on demand and returns when ready.
                If False, finds the prompt templates eagerly and returns all prompt templates.

        Returns:
            List[:class:`~bodhilib.PromptTemplate`] | Iterator[:class:`~bodhilib.PromptTemplate`]: list or iterator of
                prompt templates matching the tags
        """

    @abc.abstractmethod
    def list_all(self, stream: bool = False) -> Union[List[PromptTemplate], Iterator[PromptTemplate]]:
        """List all prompt templates in the source.

        Args:
            stream (Optional[bool]): option to stream the prompt templates.
                If True, loads the prompt templates lazily on demand and returns when queried.
                If False, lists the prompt templates eagerly and returns all prompt templates.

        Returns:
            Iterator[PromptTemplate]: list iterator of all prompt templates in the source
        """


# endregion
# region data loader
#######################################################################################################################
class DataLoader(Iterable[Document], abc.ABC):
    """Abstract base class for data loaders.

    A data loader should inherit from this class and implement the abstract methods.
    """

    @abc.abstractmethod
    def add_resource(self, **kwargs: Dict[str, Any]) -> None:
        """Add a resource to the data loader."""

    @abc.abstractmethod
    def __iter__(self) -> Iterator[Document]:
        """Returns the document iterator.

        It is for the sub-class to ensure the `__iter__` method returns a new instance of iterator
        """

    def load(self, stream: bool = False) -> Union[List[Document], Iterator[Document]]:
        """Returns the document as list or list iterator.

        Args:
            stream (Optional[bool]): option to stream the document as they load.
                If True, loads the document lazily on demand and returns when ready.
                If False, loads the documents eagerly and returns all documents.
        """
        if stream:
            return iter(self)
        return list(self)


# endregion
# region splitter
#######################################################################################################################
class Splitter(abc.ABC):
    """Splitter defines abstract method to split longer text into shorter text.

    Splitter takes in longer text as a generic :data:`~bodhilib.TextLike`
    and splits them into shorter text and return as :class:`~bodhilib.Node`.
    The shorter text are then used to create embeddings.
    """

    @abc.abstractmethod
    def split(self, inputs: SerializedInput, stream: bool = False) -> Union[List[Node], Iterator[Node]]:
        """Split a :data:`~bodhilib.SerializedInput` into a list of :class:`~bodhilib.Node`.

        Args:
            inputs (:data:`~bodhilib.SerializedInput`): takes input as :data:`~bodhilib.SerializedInput`,
                a generic type that can be a :data:`~bodhilib.TextLike`, a list of :data:`~bodhilib.TextLike`,
                or a serialized dict of the object.
            stream (Optional[bool]): option to stream the splits as they are ready.
                If True, splits the document lazily on demand and returns when ready.
                If False, splits the document eagerly and returns all splits.

        Returns:
            List[:class:`~bodhilib.Node`]: a list of :class:`~bodhilib.Node` as result of the split
        """


# endregion
# region embedder
#######################################################################################################################
class Embedder(abc.ABC):
    """Abstract base class for embedders.

    An embedder should inherit from this class and implement the abstract methods.
    """

    @abc.abstractmethod
    def embed(self, inputs: SerializedInput, stream: bool = False) -> Union[List[Node], Iterator[Node]]:
        """Embed a :data:`~bodhilib.SerializedInput` using the embedder service.

        Args:
            inputs (:data:`~bodhilib.SerializedInput`): takes input as :data:`~bodhilib.SerializedInput`,
                a generic type that can be a :data:`~bodhilib.TextLike`, a list of :data:`~bodhilib.TextLike`,
                or a serialized dict of the object.
            stream (Optional[bool]): option to stream the embeddings as they are ready.
                If True, embeds the document lazily on demand and returns when ready.
                If False, embeds the document eagerly and returns all embeddings.

        Returns:
            List[:data:`~bodhilib.Node`] | Iterator[:data:`~bodhilib.Node`]: list or iterator of :data:`~bodhilib.Node`
                enriched with :data:`~bodhilib.Embedding`
        """

    @property
    @abc.abstractmethod
    def dimension(self) -> int:
        """Dimension of the embeddings."""


# endregion
# region llm
#######################################################################################################################
class LLM(abc.ABC):
    """Abstract Base Class LLM defines the common interface implemented by all LLM implementations."""

    @abc.abstractmethod
    def generate(
        self,
        prompt_input: SerializedInput,
        *,
        stream: Optional[bool] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: Optional[int] = None,
        stop: Optional[List[str]] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Union[Prompt, PromptStream]:
        """Base class :func:`~bodhilib.LLM.generate` method interface common to all LLM service implementation.

        Takes in :data:`bodhilib.SerializedInput`, a flexible input supporting from plain string,
        or a :class:`~bodhilib.Prompt` object, or a dict representation of Prompt.
        Returns the response from LLM service as :class:`~bodhilib.Prompt` object with `source="output"`.

        Args:
            prompts (:data:`bodhilib.SerializedInput`): input to the LLM service
            stream (bool): whether to stream the response from the LLM service
            temperature (Optional[float]): temperature or randomness of the generation
            top_p (Optional[float]): token consideration probability top_p for the generation
            top_k (Optional[int]): token consideration number top_k for the generation
            n (Optional[int]): number of responses to generate
            stop (Optional[List[str]]): list of stop tokens to stop the generation
            max_tokens (Optional[int]): maximum number of tokens to generate
            presence_penalty (Optional[float]): presence penalty for the generation, between -2 and 2
            frequency_penalty (Optional[float]): frequency penalty for the generation, between -2 and 2
            user (Optional[str]): user making the request, for monitoring purpose
            kwargs (Dict[str, Any]): pass through arguments for the LLM service

        Returns:
            :class:`~bodhilib.Prompt`: a Prompt object, if stream is False
            Iterator[:class:`~bodhilib.Prompt`]: an iterator of Prompt objects, if stream is True
        """


# endregion
# region vector db
#######################################################################################################################
class VectorDBError(Exception):
    """Error to wrap any exception raised by the vector database client."""


class VectorDB(abc.ABC):
    """VectorDB defines the interface for a vector database client."""

    @abc.abstractmethod
    def ping(self) -> bool:
        """Check the connection to the vector db.

        Returns:
            True if the database is up and running.

        Raises:
            bodhilib.VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """

    @abc.abstractmethod
    def connect(self) -> bool:
        """Establishes a connection with the vector db.

        Returns:
            True if the connection with vector db is successful.

        Raises:
            bodhilib.VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """

    @abc.abstractmethod
    def close(self) -> bool:
        """Closes the connection with the vector db.

        Returns:
            True if the connection with vector db is closed successfully.

        Raises:
            bodhilib.VectorDBError: Wraps any connection error raised by the underlying database and raises.
        """

    @abc.abstractmethod
    def get_collections(self) -> List[str]:
        """List all the vector databases.

        Returns:
            List of vector database names.

        Raises:
            bodhilib.VectorDBError: Wraps any database list error raised by the underlying database.
        """

    @abc.abstractmethod
    def create_collection(
        self,
        collection_name: str,
        dimension: int,
        distance: Union[str, Distance],
        **kwargs: Dict[str, Any],
    ) -> bool:
        """Create a collection in vector database.

        Returns:
            True if the database is created successfully.

        Raises:
            VectorDBError: Wraps any database create error raised by the underlying database.
        """

    @abc.abstractmethod
    def delete_collection(self, collection_name: str, **kwargs: Dict[str, Any]) -> bool:
        """Deletes a collection in vector database.

        Returns:
            True if the database is deleted successfully.

        Raises:
            VectorDBError: Wraps any database delete error raised by the underlying client.
        """

    @abc.abstractmethod
    def upsert(self, collection_name: str, nodes: List[Node]) -> List[Node]:
        """Insert or update a node into the database along with metadata.

        This updates the nodes with record ids and other DB information, similar to using an ORM.

        Returns:
            List of record ids for the inserted embeddings.

        Raises:
            VectorDBError: Wraps any database delete error raised by the underlying client.
        """

    @abc.abstractmethod
    def query(
        self,
        collection_name: str,
        embedding: Union[Embedding, SupportsEmbedding],
        filter: Optional[Union[Dict[str, Any], Filter]] = None,
        **kwargs: Dict[str, Any],
    ) -> List[Node]:
        """Search for the nearest vectors in the database.

        Args:
            collection_name (str): name of the collection
            embedding (Embedding): embedding to search for
            filter (Optional[Union[Dict[str, Any], :class:`~bodhilib.Filter`]]): filter to apply on
                the metadata of the record
            **kwargs (Dict[str, Any]): pass through arguments for the vector db

        Returns:
            List of nodes with metadata.

        Raises:
            VectorDBError: Wraps any database delete error raised by the underlying client.
        """


# endregion
# region plugin
#######################################################################################################################
# PromptSource
PS = TypeVar("PS", bound=PromptSource)
"""TypeVar for PromptSource."""


def get_prompt_source(
    service_name: str,
    *,
    oftype: Optional[Type[PS]] = None,
    publisher: Optional[str] = "bodhiext",
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> PS:
    """Get an instance of PromptSource for given arguments."""
    if oftype is None:
        return_type: Type[Any] = PromptSource
    else:
        return_type = oftype

    manager = PluginManager.instance()
    prompt_source: PS = manager.get(
        service_name=service_name,
        service_type="prompt_source",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(PS, prompt_source)


# DataLoader
DL = TypeVar("DL", bound=DataLoader)
"""TypeVar for DataLoader."""


def get_data_loader(
    service_name: str,
    *,
    oftype: Optional[Type[DL]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> DL:
    """Get an instance of data loader for given arguments.

    Given the service name, publisher (optional) and version(optional),
    return the registered data loader oftype (optional).

    Args:
        service_name (str): name of the service, e.g. "file", "notion", "s3"
        oftype (Optional[Type[T]]): if the type of data loader is known, pass the type in argument `oftype`,
            the data loader is cast to `oftype` and returned for better IDE support.
        publisher (Optional[str]): publisher or developer of the data loader plugin, e.g. "bodhilib","<github-username>"
        version (Optional[str]): version of the data loader
        **kwargs (Dict[str, Any]): pass through arguments for the data loader, e.g. aws_access_key_id, notion_db etc.

    Returns:
        DL (:data:`~bodhilib.DL` | :class:`~bodhilib.DataLoader`): an instance of DataLoader service
            of type `oftype`, if oftype is passed, else of type :class:`~bodhilib.DataLoader`

    Raises:
        TypeError: if the type of data loader is not oftype
    """
    if oftype is None:
        return_type: Type[Any] = DataLoader
    else:
        return_type = oftype

    manager = PluginManager.instance()
    data_loader: DL = manager.get(
        service_name=service_name,
        service_type="data_loader",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(DL, data_loader)


def list_data_loaders() -> List[Service]:
    """List all data loaders installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("data_loader")


# Splitter
S = TypeVar("S", bound=Splitter)
"""TypeVar for Splitter."""


def get_splitter(
    service_name: str,
    *,
    oftype: Optional[Type[S]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> S:
    """Get an instance of splitter for given arguments.

    Returns:
        S (:data:`~bodhilib.S` | :class:`~bodhilib.Splitter`): an instance of Splitter service
    """
    if oftype is None:
        return_type: Type[Any] = Splitter
    else:
        return_type = oftype

    manager = PluginManager.instance()
    splitter: S = manager.get(
        service_name=service_name,
        service_type="splitter",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(S, splitter)


# Embedder
E = TypeVar("E", bound=Embedder)
"""TypeVar for Embedder."""


def get_embedder(
    service_name: str,
    *,
    oftype: Optional[Type[E]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> E:
    """Get an instance of embedder given the service name, publisher (optional) and version(optional).

    Args:
        service_name (str): name of the service, e.g. "sentence-transformers" etc.
        oftype (Optional[Type[T]]): if the type of embedder is known, pass the type in argument `oftype`,
            the embedder is cast to `oftype` and returned for better IDE support.
        publisher (Optional[str]): publisher or developer of the embedder plugin, e.g. "bodhilib","<github-username>"
        version (Optional[str]): version of the embedder
        **kwargs (Dict[str, Any]): pass through arguments for the embedder, e.g. dimension etc.

    Returns:
        E (:data:`~bodhilib.E` | :class:`~bodhilib.Embedder`): an instance of Embedder service
            of type `oftype`, if oftype is passed, else of type :class:`~bodhilib.Embedder`

    Raises:
        TypeError: if the type of embedder is not oftype
    """
    if oftype is None:
        return_type: Type[Any] = Embedder
    else:
        return_type = oftype

    manager = PluginManager.instance()
    embedder: E = manager.get(
        service_name, "embedder", oftype=return_type, publisher=publisher, version=version, **kwargs
    )
    return cast(E, embedder)


def list_embedders() -> List[Service]:
    """List all embedders installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("embedder")


# LLM
L = TypeVar("L", bound=LLM)
"""TypeVar for LLM."""


def get_llm(
    service_name: str,
    model: str,
    api_key: Optional[str] = None,
    *,
    oftype: Optional[Type[L]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> L:
    """Get an instance of LLM for the given service name and model.

    Args:
        service_name (str): name of the service, e.g. "openai", "cohere", "anthropic"
        model (str): name of the model, e.g. "chat-gpt-3.5", "command", "claude-2"
        api_key (Optional[str]): API key for the service, if the api_key is not provided,
            it will be looked in the environment variables
        oftype (Optional[Type[T]]): if the type of LLM is known, pass the type in argument `oftype`,
            the LLM is cast to `oftype` and returned for better IDE support.
        publisher (Optional[str]): publisher or developer of the service plugin, e.g. "bodhilib", "<github-username>"
        version (Optional[str]): version of the service
        **kwargs (Dict[str, Any]): pass through arguments for the LLM service, e.g. "temperature", "max_tokens", etc.

    Returns:
        T (:data:`~bodhilib.L` | :class:`~bodhilib.LLM`):
            an instance of LLM service of type `oftype`, if oftype is passed,
            else of type :class:`~bodhilib.LLM`

    Raises:
        TypeError: if the type of LLM is not oftype
    """
    if oftype is None:
        return_type: Type[Any] = LLM
    else:
        return_type = oftype

    manager = PluginManager.instance()
    llm: L = manager.get(
        service_name=service_name,
        service_type="llm",
        oftype=return_type,
        publisher=publisher,
        version=version,
        model=model,
        api_key=api_key,
        **kwargs,
    )
    return cast(L, llm)


def list_llms() -> List[Service]:
    """List all LLM services installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("llm")


# VectorDB
V = TypeVar("V", bound=VectorDB)
"""TypeVar for VectorDB."""


def get_vector_db(
    service_name: str,
    *,
    oftype: Optional[Type[V]] = None,
    publisher: Optional[str] = None,
    version: Optional[str] = None,
    **kwargs: Dict[str, Any],
) -> V:
    """Get an instance of VectorDB for the given service name.

    Returns:
        V (:data:`~bodhilib.V` | :class:`~bodhilib.VectorDB`): an instance of VectorDB service
    """
    if oftype is None:
        return_type: Type[Any] = VectorDB
    else:
        return_type = oftype

    manager = PluginManager.instance()
    vector_db: V = manager.get(
        service_name=service_name,
        service_type="vector_db",
        oftype=return_type,
        publisher=publisher,
        version=version,
        **kwargs,
    )
    return cast(V, vector_db)


def list_vector_dbs() -> List[Service]:
    """List all VectorDB services installed and available."""
    manager = PluginManager.instance()
    return manager.list_services("vector_db")


# endregion



# this file is automatically generated and overwritten by CI process
# to modify the content, modify the CI workflow file - .github/workflows/main.yml, job: "Bump patch version"
__version__ = "0.1.11"



from __future__ import annotations

import logging
import os


def init_logger(package_name: str) -> logging.Logger:
    """Initialize logger for bodhilib.

    Returns the default logger if BODHILIB_LOG_LEVEL is not set, else returns a logger with
    the name "bodhilib" and log level set to BODHILIB_LOG_LEVEL. Optionally also sets the
    format of the log message using BODHILIB_LOG_FORMAT.

    Returns:
        logging.Logger: logger for bodhilib
    """
    # if library logging level not set, set the logger as the root logger
    log_level = os.environ.get("BODHILIB_LOG_LEVEL", None)
    if not log_level:
        return logging.getLogger()
    logger = logging.getLogger(package_name)
    logger.setLevel(log_level)
    handler = logging.StreamHandler()
    format = os.environ.get("BODHILIB_LOG_FORMAT", "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    formatter = logging.Formatter(format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



from __future__ import annotations

import io
import itertools
import reprlib
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    no_type_check,
)

from pydantic import BaseModel, Field, validator
from typing_extensions import TypeAlias

# region type aliases
#######################################################################################################################
# Update the documentation in bodhilib.rst file directly. TypeAlias inline documentation not picked up by sphinx.
# Duplicating it here to make it easier when browsing the code.
PathLike: TypeAlias = Union[str, Path]
"""Type alias for Union of :class:`str` and :class:`~pathlib.Path`"""
TextLike: TypeAlias = Union[str, "SupportsText"]
"""Type alias for Union of :class:`str` and protocol :data:`~bodhilib.SupportsText`"""
TextLikeOrTextLikeList: TypeAlias = Union[TextLike, Iterable[TextLike]]
"""Type alias for Union of :data:`~bodhilib.TextLike` or list of :data:`~bodhilib.TextLike`"""
SerializedInput: TypeAlias = Union[TextLikeOrTextLikeList, Dict[str, Any], Iterable[Dict[str, Any]]]
"""Type alias for various inputs that can be passed to the components."""
Embedding: TypeAlias = List[float]
"""Type alias for list of :class:`float`, to indicate the embedding generated
from :class:`~bodhilib.Embedder` operation"""


class SupportsText(Protocol):
    """TextLike is a protocol for types that can be converted to text.

    To support the protocol, the type must have a property `text`.

    Known sub-classes: :class:`~bodhilib.Prompt`, :class:`~bodhilib.Document`, :class:`~bodhilib.Node`
    """

    @property
    def text(self) -> str:
        """Return the content of the object as string."""


class SupportsEmbedding(Protocol):
    """SupportsEmbedding matches the types that contains a field `embedding` of type :data:`~bodhilib.Embedding`."""

    @property
    def embedding(self) -> Embedding:
        """Return the embedding of the object as :data:`~bodhilib.Embedding`."""


def supportstext(obj: object) -> bool:
    """Returns True if the object supports :data:`~bodhilib.SupportsText` protocol."""
    return hasattr(obj, "text")


def istextlike(obj: object) -> bool:
    """Returns True if the object is a :data:`~TextLike`."""
    return isinstance(obj, str) or supportstext(obj)


def supportsembedding(obj: object) -> bool:
    """Returns True if the object supports :data:`~bodhilib.SupportsEmbedding` protocol."""
    return hasattr(obj, "embedding")


# endregion
# region utility
#######################################################################################################################
class _StrEnumMixin:
    """Mixin class for string enums, provides __str__ and __eq__ methods."""

    @no_type_check
    def __str__(self) -> str:
        """Returns the string value of the string enum."""
        return self.value

    @no_type_check
    def __eq__(self, other: Any) -> bool:
        """Compares this string enum to other string enum or string values."""
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, type(self)):
            return self.value == other.value
        return False

    @no_type_check
    @classmethod
    def membersstr(cls) -> List[str]:
        return [e.value for e in cls.__members__.values()]


_EnumT = TypeVar("_EnumT", bound=Enum)
"""TypeVar for Enum type."""


def _strenum_validator(enum_cls: Type[_EnumT], value: Any) -> _EnumT:
    """Converts a string value to an enum value."""
    if isinstance(value, str):
        try:
            return enum_cls[value.upper()]
        except KeyError as e:
            allowed_values = [e.value for e in enum_cls]
            raise ValueError(f"Invalid value for {enum_cls.__name__}. Allowed values are {allowed_values}.") from e
    elif isinstance(value, enum_cls):
        return value
    else:
        raise ValueError(f"Invalid type for value, {type(value)=}")


# endregion
# region value objects
#######################################################################################################################
class Role(_StrEnumMixin, str, Enum):
    """Role of the prompt.

    Used for fine-grain control over "role" instructions to the LLM service.
    Can be one of - *'system', 'ai', or 'user'*.
    """

    SYSTEM = "system"
    AI = "ai"
    USER = "user"


class Source(_StrEnumMixin, str, Enum):
    """Source of the prompt.

    If the prompt is given as input by the user, then *source="input"*,
    or if the prompt is generated as response by the LLM service, then *source="output"*.
    """

    INPUT = "input"
    OUTPUT = "output"


class Distance(_StrEnumMixin, str, Enum):
    """Vector Distance Method."""

    COSINE = "cosine"
    EUCLIDEAN = "euclidean"
    DOT_PRODUCT = "dot_product"


# endregion
# region prompt
#######################################################################################################################
class Prompt(BaseModel):
    """Prompt encapsulating input/output schema to interact with LLM service."""

    role: Role = Role.USER
    """The role of the prompt.

    Defaults to :obj:`Role.USER`."""

    source: Source = Source.INPUT
    """The source of the prompt.

    Defaults to :obj:`Source.INPUT`."""

    text: str
    """The text or content or input component of the prompt."""

    class Config:
        use_enum_values = True

    # overriding __init__ to provide positional argument construction for prompt. E.g. `Prompt("text")`
    def __init__(
        self,
        text: str,
        role: Optional[Union[Role, str]] = Role.USER,
        source: Optional[Union[Source, str]] = Source.INPUT,
    ):
        """Initialize a prompt.

        Args:
            text (str): text of the prompt
            role (Role): role of the prompt.

                Role can be given as one of the allowed string value ["system", "ai", "user"]
                or as a Role enum [:obj:`Role.SYSTEM`, :obj:`Role.AI`, :obj:`Role.USER`].

                The string is converted to Role enum. If the string value is not one of the allowed values,
                then a ValueError is raised.
            source (Source): source of the prompt.

                Source can be given as a one of the allowed string value ["input", "output"]
                or as a Source enum [:obj:`Source.INPUT`, :obj:`Source.OUTPUT`].

                The string is converted to Source enum. If the string value is not one of the allowed values,
                then a ValueError is raised.

        Raises:
            ValueError: If the role or source is not one of the allowed values.
        """
        role = role or Role.USER
        source = source or Source.INPUT
        super().__init__(text=text, role=role, source=source)

    @validator("role", pre=True, always=True)
    def validate_role(cls, value: Any) -> Role:
        return _strenum_validator(Role, value)

    @validator("source", pre=True, always=True)
    def validate_source(cls, value: Any) -> Source:
        return _strenum_validator(Source, value)

    def isstream(self) -> bool:
        """To check if this is a prompt stream.

        Returns:
            bool: False as this is not a prompt stream.
        """
        return False


T = TypeVar("T")
"""TypeVar for LLM Response Type."""


class PromptStream(Iterator[Prompt]):
    """Iterator over a stream of prompts.

    Used by LLMs to wrap the stream response to an iterable over prompts.
    """

    def __init__(self, api_response: Iterable[T], transformer: Callable[[T], Prompt]):
        """Initialize a prompt stream.

        Args:
            api_response (Iterable[T]): LLM API Response of generic type :data:`~bodhilib.T` as Iterable
            transformer (Callable[[T], Prompt]): Transformer function to convert API response to Prompt
        """
        self.api_response = iter(api_response)
        self.transformer = transformer
        self.output = io.StringIO()
        self.role: Optional[str] = None

    def __iter__(self) -> Iterator[Prompt]:
        """Returns the iterator object itself."""
        return self

    def __next__(self) -> Prompt:
        """Returns the next item from the iterator as Prompt object."""
        try:
            chunk_response = next(self.api_response)
        except StopIteration as e:
            raise StopIteration from e
        prompt = self.transformer(chunk_response)
        if self.role is None:
            self.role = prompt.role
        self.output.write(prompt.text)
        return prompt

    def isstream(self) -> bool:
        """To check if this is a prompt stream.

        Returns:
            bool: False as this is not a prompt stream.
        """
        return True

    @property
    def text(self) -> str:
        # TODO change it to first read all the stream content, and then return the text
        # Create another property to return so-far accumulated text
        """Returns the text accumulated over the stream of responses."""
        return self.output.getvalue()


def prompt_user(text: str) -> Prompt:
    """Factory method to generate user prompt from string.

    Args:
        text: text of the prompt

    Returns:
        Prompt: Prompt object generated from the text. Defaults role="user" and source="input".
    """
    return Prompt(text=text, role=Role.USER, source=Source.INPUT)


def prompt_system(text: str) -> Prompt:
    """Factory method to generate system prompt from string.

    Args:
        text: text of the prompt

    Returns:
        Prompt: Prompt object generated from the text. Defaults role="system" and source="input".
    """
    return Prompt(text=text, role=Role.SYSTEM, source=Source.INPUT)


def prompt_output(text: str) -> Prompt:
    """Factory method to generate output prompts.

    Generates a prompt with source="output". Mainly by LLMs to generate output prompts.
    """
    return Prompt(text=text, role=Role.AI, source=Source.OUTPUT)


# endregion
# region document
#######################################################################################################################
class Document(BaseModel):
    """Document defines the basic interface for a processible resource.

    Primarily contains text (content) and metadata.
    """

    text: str
    """Text content of the document."""

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata associated with the document. e.g. filename, dirname, url etc."""

    def __repr__(self) -> str:
        """Returns a string representation of the document."""
        return f"Document(text={reprlib.repr(self.text)}, metadata={reprlib.repr(self.metadata)})"


class Node(BaseModel):
    """Node defines the basic data structure for a processible resource.

    It contains a unique identifier, content text, metadata associated with its sources,
    and embeddings.
    """

    id: Optional[str] = None
    """Unique identifier for the node.

    Generated during the document split operation, or retrieved from doc/vector database at the time of query."""

    text: str
    """Text content of the document."""

    parent: Optional[Document] = None
    """Metadata associated with the document. e.g. filename, dirname, url etc."""

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata associated with the node. This is also copied over from parent when splitting Document."""

    embedding: Optional[Embedding] = None

    def __repr__(self) -> str:
        """Returns a string representation of the document."""
        return f"Node(id={self.id}, text={reprlib.repr(self.text)}, parent={repr(self.parent)})"


# endregion
# region model converters
#######################################################################################################################
def to_document(textlike: TextLike) -> Document:
    """Converts a :data:`~bodhilib.TextLike` to :class:`~bodhilib.Document`."""
    if isinstance(textlike, Document):
        return textlike
    elif isinstance(textlike, str):
        return Document(text=textlike)
    elif supportstext(textlike):
        return Document(text=textlike.text)
    raise ValueError(f"Cannot convert type {type(textlike)} to Document.")


def to_prompt(textlike: TextLike) -> Prompt:
    """Converts a :data:`~TextLike` to :class:`~Prompt`."""
    if isinstance(textlike, Prompt):
        return textlike
    elif isinstance(textlike, str):
        return Prompt(text=textlike)
    elif supportstext(textlike):
        return Prompt(text=textlike.text)
    raise ValueError(f"Cannot convert type {type(textlike)} to Prompt.")


def to_node(textlike: TextLike) -> Node:
    """Converts a :data:`~TextLike` to :class:`~Node`."""
    if isinstance(textlike, Node):
        return textlike
    elif isinstance(textlike, str):
        return Node(text=textlike)
    elif supportstext(textlike):
        return Node(text=textlike.text)
    raise ValueError(f"Cannot convert type {type(textlike)} to Node.")


def to_text(textlike: TextLike) -> str:
    """Converts a :data:`~TextLike` to string."""
    if isinstance(textlike, str):
        return textlike
    if supportstext(textlike):
        return textlike.text
    raise ValueError(f"Cannot convert type {type(textlike)} to text.")


def to_prompt_list(inputs: SerializedInput) -> List[Prompt]:
    """Converts a :data:`~bodhilib.SerializedInput` to list of :class:`~Prompt`."""
    if istextlike(inputs):
        return [to_prompt(cast(TextLike, inputs))]  # cast to fix mypy warning
    elif isinstance(inputs, dict):
        return [Prompt(**inputs)]
    elif isinstance(inputs, Iterable):
        result = [to_prompt_list(textlike) for textlike in inputs]
        return list(itertools.chain(*result))
    else:
        return [to_prompt(inputs)]


def to_document_list(inputs: SerializedInput) -> List[Document]:
    """Converts a :data:`~bodhilib.SerializedInput` to list of :class:`~Document`."""
    if istextlike(inputs):
        return [to_document(cast(TextLike, inputs))]  # cast to fix mypy warning
    elif isinstance(inputs, dict):
        return [Document(**inputs)]
    if isinstance(inputs, Iterable):
        result = [to_document_list(input) for input in inputs]
        return list(itertools.chain(*result))
    else:
        return [to_document(inputs)]


def to_node_list(inputs: SerializedInput) -> List[Node]:
    """Converts a :data:`~bodhilib.SerializedInput` to list of :class:`~Node`."""
    if (
        not isinstance(inputs, BaseModel)  # BaseModel is Iterable
        and isinstance(inputs, Iterable)  # if is list
        and all(isinstance(input, Node) for input in inputs)  # and if all are Node instance
    ):
        return inputs  # type: ignore
    if istextlike(inputs):
        return [to_node(cast(TextLike, inputs))]  # cast to fix mypy warning
    elif isinstance(inputs, dict):
        return [Node(**inputs)]
    elif isinstance(inputs, Iterable):
        result = [to_node_list(input) for input in inputs]
        return list(itertools.chain(*result))
    else:
        return [to_node(inputs)]


# endregion
# region utils
#######################################################################################################################


# endregion



"""Logging module for bodhilib."""

from bodhilib.common import package_name

from ._logging import init_logger

#: logging.Logger: logger used by bodhilib and plugins for logging
logger = init_logger(package_name)

__all__ = ["logger"]



""":mod:`bodhilib` module defines classes and methods for bodhilib components."""
import inspect

from ._components import LLM as LLM
from ._components import DataLoader as DataLoader
from ._components import Embedder as Embedder
from ._components import PromptSource as PromptSource
from ._components import PromptTemplate as PromptTemplate
from ._components import Splitter as Splitter
from ._components import VectorDB as VectorDB
from ._components import VectorDBError as VectorDBError
from ._components import get_data_loader as get_data_loader
from ._components import get_embedder as get_embedder
from ._components import get_llm as get_llm
from ._components import get_prompt_source as get_prompt_source
from ._components import get_splitter as get_splitter
from ._components import get_vector_db as get_vector_db
from ._components import list_embedders as list_embedders
from ._components import list_llms as list_llms
from ._components import list_vector_dbs as list_vector_dbs
from ._filter import And as And
from ._filter import Condition as Condition
from ._filter import Filter as Filter
from ._filter import Nor as Nor
from ._filter import OperatorCondition as OperatorCondition
from ._filter import Or as Or
from ._models import Distance as Distance
from ._models import Document as Document
from ._models import Embedding as Embedding
from ._models import Node as Node
from ._models import PathLike as PathLike
from ._models import Prompt as Prompt
from ._models import PromptStream as PromptStream
from ._models import Role as Role
from ._models import SerializedInput as SerializedInput
from ._models import Source as Source
from ._models import SupportsEmbedding as SupportsEmbedding
from ._models import SupportsText as SupportsText
from ._models import TextLike as TextLike
from ._models import TextLikeOrTextLikeList as TextLikeOrTextLikeList
from ._models import _StrEnumMixin as _StrEnumMixin
from ._models import istextlike as istextlike
from ._models import prompt_output as prompt_output
from ._models import prompt_system as prompt_system
from ._models import prompt_user as prompt_user
from ._models import supportsembedding as supportsembedding
from ._models import supportstext as supportstext
from ._models import to_document as to_document
from ._models import to_document_list as to_document_list
from ._models import to_node as to_node
from ._models import to_node_list as to_node_list
from ._models import to_prompt as to_prompt
from ._models import to_prompt_list as to_prompt_list
from ._models import to_text as to_text
from ._plugin import PluginManager as PluginManager
from ._plugin import Service as Service
from ._plugin import service_provider as service_provider
from ._version import __version__ as __version__
from ._yaml import yaml_dump as yaml_dump
from ._yaml import yaml_load as yaml_load
from .common import package_name as package_name

__all__ = [name for name, obj in globals().items() if not (name.startswith("_") or inspect.ismodule(obj))]

del inspect



from __future__ import annotations

import itertools
import sys
from typing import Any, Callable, List, NamedTuple, Optional, Type, TypeVar, cast

import pluggy

from bodhilib.common import package_name
from bodhilib.logging import logger

hookspec = pluggy.HookspecMarker(package_name)
service_provider = pluggy.HookimplMarker(package_name)
current_module = sys.modules[__name__]


class Service(NamedTuple):
    """Encapsulates basic info of service provided by the plugin."""

    service_name: str
    """Name of the service provided by the plugin. E.g. openai, cohere, anthropic, etc."""
    service_type: str
    """Type of service provided by the plugin. E.g. llm, embed etc."""
    publisher: str
    """Publisher identifier of the plugin. E.g. bodhilib, openai, <github-user> etc."""
    service_builder: Callable  # signature(**kwargs: Dict[str, Any])
    """Callable which returns an instance of service.

    Callable takes in `**kwargs: Dict[str, Any]` and returns an instance of service.
    """
    version: str = ""
    """Version of the plugin"""


@hookspec
def bodhilib_list_services() -> List[Service]:
    """Return a list of services supported by plugin.

    When user request for service using service_name and service_type, this information is matched,
    and the service_builder method of the plugin is called with all info.

    Returns:
        List[Service]: list of services supported by plugin
    """
    return []


C = TypeVar("C")
"""TypeVar for Component (one of sub-class of :class:`~bodhilib.LLM`, :class:`~bodhilib.Embedder`,
:class:`~bodhilib.DataLoader`).
Used for type hinting in :meth:`~bodhilib.plugin.PluginManager.get` method."""


class PluginManager:
    """Searches for and loads bodhilib plugins."""

    _instance = None

    def __new__(cls) -> "PluginManager":
        """Override `__new__` in case constructor is directly called."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def instance(cls) -> "PluginManager":
        """Singleton method to return instance of PluginManager."""
        # extracting in variable because of mypy warning
        instance = cls._instance
        if instance is None:
            return cls()
        return instance

    def __init__(self) -> None:
        """Initialize plugin manager and load the bodhilib plugins."""
        pm = pluggy.PluginManager(package_name)
        pm.add_hookspecs(current_module)
        pm.load_setuptools_entrypoints(package_name)
        self.pm = pm
        self.services: Optional[List[Service]] = None

    def get(
        self,
        service_name: str,
        service_type: str,
        *,
        oftype: Optional[Type[C]] = None,
        publisher: Optional[str] = None,
        version: Optional[str] = None,
        **kwargs: Any,
    ) -> C:
        """Get an instance of service for the given service and type.

        Args:
            service_name (str): name of the service, e.g. "openai", "cohere", "anthropic"
            service_type (str): type of the service, e.g. "llm", "embedder", "data_loader"
            oftype (Optional[Type[T]]): if the type of service is known, pass the type in argument `oftype`,
                the service is cast to `oftype` and returned for better IDE support.
            publisher (Optional[str]): publisher or developer of the service plugin, e.g. "bodhilib","<github-username>"
            version (Optional[str]): version of the service
            **kwargs (Dict[str, Any]): pass through arguments for the service, e.g. "temperature", "max_tokens", etc.

        Returns:
            C (:data:`~bodhilib.C` | :class:`~typing.Any`):
                an instance of service of type `oftype`, if oftype is passed, else of type :class:`~typing.Any`

        Raises:
            TypeError: if the type of service is not oftype
        """
        if oftype is None:
            return_type: Type[Any] = type(Any)
        else:
            return_type = oftype
        if self.services is None:
            self.services = self._fetch_services()
        all_args = {
            "service_name": service_name,
            "service_type": service_type,
            "publisher": publisher,
            "version": version,
            **kwargs,
        }
        # remove None values
        all_args = {k: v for k, v in all_args.items() if v is not None}
        for service in self.services:
            if service.service_name == service_name and service.service_type == service_type:
                component = service.service_builder(**all_args)
                if not isinstance(component, return_type):
                    raise TypeError(f'Expecting {service_type} of type "{oftype}", but got "{type(component)}"')
                return cast(C, component)
        raise ValueError(f"Service {service_name=} of type {service_type=} not found in registered services")

    def list_services(self, service_type: str) -> List[Service]:
        """List all services of type service_type installed and available."""
        if self.services is None:
            self.services = self._fetch_services()
        return [s for s in self.services if s.service_type == service_type]

    def _fetch_services(self) -> List[Service]:
        logger.debug({"msg": "fetching services"})
        services = list(itertools.chain(*self.pm.hook.bodhilib_list_services()))
        logger.debug({"msg": "fetched services", "services": services})
        # get list of services which are not instance of Service and log with warning
        invalid_services = [p for p in services if not isinstance(p, Service)]
        if invalid_services:
            logger.warning({"msg": "invalid services, ignoring", "services": invalid_services})
        # get list of valid services and log with debug
        valid_services = [p for p in services if isinstance(p, Service)]
        logger.debug({"msg": "valid services", "services": valid_services})
        return valid_services



from typing import Any, Dict, List, Protocol


class Condition(Protocol):
    """Represents the filter expression of MongoDB query expression."""

    def evaluate(self, record: Dict[str, Any]) -> bool:
        """Returns the result after applying the condition to the record."""

    def to_dict(self) -> Dict[str, Any]:
        """Returns the condition as a dict."""


class OperatorCondition:
    """Represents the filter expression with an operator of MongoDB query expression."""

    operators: List[str] = ["$eq", "$gt", "$lt", "$ne", "$gte", "$lte", "$in", "$nin", "$all"]

    def __init__(self, field: str, operator: str, value: Any) -> None:
        """Construct the Operator condition with given arguments.

        Args:
            field (str): The field name.
            operator (str): The operator.
                One of "$eq", "$gt", "$lt", "$ne", "$gte", "$lte", "$in", "$nin", "$all".
            value (Any): The value to compare.

        Raises:
            ValueError: If the operator is invalid.
        """
        if operator not in self.operators:
            raise ValueError(f"Invalid operator: {operator}")
        self.field: str = field
        self.operator: str = operator
        self.value: Any = value

    def evaluate(self, record: Dict[str, Any]) -> bool:
        """Return the predicate for applying the Operator condition."""
        if self.field not in record:
            raise ValueError(f"field with name {self.field} not found in record, found {record.keys()}")
        if self.operator == "$eq":
            return record[self.field] == self.value  # type: ignore[no-any-return]
        elif self.operator == "$gt":
            return record[self.field] > self.value  # type: ignore[no-any-return]
        elif self.operator == "$lt":
            return record[self.field] < self.value  # type: ignore[no-any-return]
        elif self.operator == "$ne":
            return record[self.field] != self.value  # type: ignore[no-any-return]
        elif self.operator == "$gte":
            return record[self.field] >= self.value  # type: ignore[no-any-return]
        elif self.operator == "$lte":
            return record[self.field] <= self.value  # type: ignore[no-any-return]
        elif self.operator in ("$in", "$nin", "$all"):
            record_tags = record.get(self.field)
            if not isinstance(record_tags, list):
                raise ValueError(f"field with name {self.field} is not a list, found {type(record_tags)}")
            if self.operator == "$in":
                return any(tag in record_tags for tag in self.value)
            elif self.operator == "$nin":
                return not any(tag in record_tags for tag in self.value)
            elif self.operator == "$all":
                return all(tag in record_tags for tag in self.value)
        raise ValueError("Invalid operator: {self.operator}")

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the Operator condition as a dict."""
        return {self.field: {self.operator: self.value}}


class And:
    """Represents the And filter of MongoDB query expression."""

    def __init__(self, conditions: List[Condition]) -> None:
        """Construct the And condition with given arguments."""
        self.conditions = conditions

    def evaluate(self, record: Dict[str, Any]) -> bool:
        """Return the predicate for applying the And condition."""
        return all(condition.evaluate(record) for condition in self.conditions)

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the And condition as a dict."""
        return {"$and": [condition.to_dict() for condition in self.conditions]}


class Or:
    """Represents the Or filter of MongoDB query expression."""

    def __init__(self, conditions: List[Condition]) -> None:
        """Construct the Or condition with given arguments."""
        self.conditions = conditions

    def evaluate(self, record: Dict[str, Any]) -> bool:
        """Return the predicate for applying the filter Or condition."""
        return any(condition.evaluate(record) for condition in self.conditions)

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the Or condition as a dict."""
        return {"$or": [condition.to_dict() for condition in self.conditions]}


class Nor:
    """Represents the Nor filter of MongoDB query expression."""

    def __init__(self, conditions: List[Condition]) -> None:
        """Construct the Nor condition with given arguments."""
        self.conditions = conditions

    def evaluate(self, record: Dict[str, Any]) -> bool:
        """Return the predicate for applying the filter Nor condition."""
        return not any(condition.evaluate(record) for condition in self.conditions)

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the Nor condition as a dict."""
        return {"$nor": [condition.to_dict() for condition in self.conditions]}


class Filter:
    """Represents the MongoDB query expression."""

    def __init__(self, filter_expr: Dict[str, Any]) -> None:
        """Construct a Filter with given arguments.

        Args:
            filter_expr (dict): The MongoDB query expression as dict.
        """
        self.condition = self._parse(filter_expr)

    @staticmethod
    def from_dict(filter_dict: Dict[str, Any]) -> "Filter":
        """Construct a Filter from a dict."""
        return Filter(filter_dict)

    def to_dict(self) -> Dict[str, Any]:
        """Serialize the Filter as a dict."""
        return self.condition.to_dict()

    def _parse(self, condition: Dict[str, Any]) -> Condition:
        if "$and" in condition:
            return And([self._parse(sub_condition) for sub_condition in condition["$and"]])
        elif "$or" in condition:
            return Or([self._parse(sub_condition) for sub_condition in condition["$or"]])
        elif "$nor" in condition:
            return Nor([self._parse(sub_condition) for sub_condition in condition["$nor"]])
        for field, value in condition.items():
            if isinstance(value, dict):
                for op, op_value in value.items():
                    return OperatorCondition(field, op, op_value)
            else:
                return OperatorCondition(field, "$eq", value)
        raise ValueError(f"Invalid condition: {condition}")

    def evaluate(self, record: Dict[str, Any]) -> bool:
        """Return the predicate for applying the filter condition."""
        return self.condition.evaluate(record)
